/**
  *  \file arch/win32/win32.cc
  *  \brief Win32 Platform Utilities
  */

#include <stdio.h>
#include <windows.h>
#include "arch/win32/win32.h"
#include "cpluslib/utf8.h"

/** Convert UTF-8 filename string to Unicode string.
    \param out [out] Unicode string
    \param str [in] String */
void
win32::convertToUnicode(WStr& out, string_t str, bool terminate)
{
    Utf8Reader rdr(str);
    while (rdr.hasMore())
        out.push_back(rdr.get());
    if (terminate)
        out.push_back(0);
}

/** Convert null-terminated unicode string to UTF-8.
    \param data [in] Unicode string */
string_t
win32::convertFromUnicode(const wchar_t* data)
{
    string_t result;
    while (*data) {
        appendUtf8(result, *data);
        ++data;
    }
    return result;
}

/** Convert UTF-8 string to any codepage. */
string_t
win32::convertToCodepage(string_t str, uint32_t codepage)
{
    // Convert to Unicode
    WStr wstr;
    convertToUnicode(wstr, str, false);

    // Convert Unicode to ANSI
    int sizeNeeded = WideCharToMultiByte(codepage, 0, &wstr[0], wstr.size(), 0, 0, 0, 0);
    std::vector<char> astr(sizeNeeded);
    WideCharToMultiByte(codepage, 0, &wstr[0], wstr.size(), &astr[0], sizeNeeded, 0, 0);
    return string_t(&astr[0], sizeNeeded);
}

/** Convert UTF-8 string to ANSI. */
string_t
win32::convertToANSI(string_t str)
{
    return convertToCodepage(str, CP_ACP);
}

/** Convert null-terminated codepage string to UTF-8.
    \param data [in] Codepage string */
string_t
win32::convertFromCodepage(const char* data, uint32_t codepage)
{
    int sizeNeeded = MultiByteToWideChar(codepage, 0, data, -1, 0, 0);
    WStr wstr(sizeNeeded+1);
    MultiByteToWideChar(codepage, 0, data, -1, &wstr[0], sizeNeeded);
    return convertFromUnicode(&wstr[0]);
}

/** Convert null-terminated ANSI string to UTF-8.
    \param data [in] ANSI string */
string_t
win32::convertFromANSI(const char* data)
{
    return convertFromCodepage(data, CP_ACP);
}

/** Get text message corresponding to last error message (GetLastError).
    This returns the locale-specific string, converted to UTF-8. */
string_t
win32::getLastErrorString()
{
    DWORD err = GetLastError();
    string_t result;
    if (hasUnicodeSupport()) {
        // Unicode version
        wchar_t* msg = 0;
        if (FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                           0, err,
                           MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                           (LPWSTR) &msg,
                           0, 0) != 0)
        {
            result = convertFromUnicode(msg);
        }
        if (msg) {
            LocalFree(msg);
        }
    } else {
        // ANSI version
        char* msg = 0;
        if (FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                           0, err,
                           MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                           (LPSTR) &msg,
                           0, 0) != 0)
        {
            result = convertFromANSI(msg);
        }
        if (msg) {
            LocalFree(msg);
        }
    }

    // We expect the message to not end with a linefeed, we add that ourselves.
    // Windows messages have CR+LF at the end.
    string_t::size_type n = result.size();
    while (n > 0 && (result[n-1] == '\n' || result[n-1] == '\r'))
        --n;
    result.erase(n);

    // Fallback if Windows doesn't have a text
    if (result.size() == 0) {
        char tmp[100];
        sprintf(tmp, "Unknown error #%d (0x%08X)", int(err), unsigned(err));
        result = tmp;
    }

    return result;
}

/** Check for Unicode support. */
bool
win32::hasUnicodeSupport()
{
    // The highest bit is 0 on Windows NT, 1 on Win95/Win32s.
    static DWORD version = GetVersion();
    return version <= 0x7FFFFFFFU;
}
