/**
  *  \file client/cscreen.cc
  *  \brief Control Screens, and Related Things
  *
  *  This contains the main classes for control screens.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/chartusr.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/playerscr.h"
#include "client/screens.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/skin.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"
#include "client/wscanchart.h"
#include "client/history.h"
#include "client/dlg-visual.h"
#include "int/intglobal.h"
#include "client/dlg-comment.h"
#include "client/dialogs/searchdlg.h"
#include "client/dialogs/consoledlg.h"
#include "int/simple.h"
#include "game/global.h"
#include "ui/prefix.h"
#include "client/widgets/selswitch.h"
#include "int/process.h"
#include "client/widgets/decaymsg.h"
#include "int/interror.h"

//#define BLUESKIN
#ifndef BLUESKIN
#  define ICOLOR (COLOR_SHIELD+3)
#  define IFILL GRAY50_FILL
#else
#  define ICOLOR (COLOR_SHIELD+1)
#  define IFILL SOLID_FILL
#  define FCOLOR (COLOR_SHIELD+5)
#endif

/************************** WControlScreen::Skin ***************************/

/** \class WControlScreen::Skin
    \brief Skin for control screens */

class WControlScreen::Skin : public UISkin {
 public:
    Skin()
        : UISkin(DARK_COLOR_SET)
        { }
    void drawBackground(GfxContext& ctx, GfxRect r);
};

#ifdef BLUESKIN
void
WControlScreen::Skin::drawBackground(GfxContext& ctx, GfxRect r)
{
    ctx.getCanvas().drawBar(r, standard_colors[ICOLOR], standard_colors[0], IFILL, GFX_ALPHA_OPAQUE);
}
#else
void
WControlScreen::Skin::drawBackground(GfxContext& ctx, GfxRect r)
{
    drawSolidBar(ctx.getCanvas(), r, standard_colors[COLOR_GRAYSCALE+5]);
}
#endif

/***************************** WControlScreen ****************************/

/*
  Layout:

    ,-----------------------------------.
    |              ,------------------. |   <-- borders are 5 pixels
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              `------------------' |  <-- tile_new_y  \ 3 lines for
    |              [                  ] |  ___ chart_y2    / scanner widget
    |                                   |
    |                                   |    8 lines free here for tiles
    |                                   |
    |                                   |
    `-----------------------------------'  ---

    |  320 pixels  |
                chart_x1
 */

/** Constructor.
    \param type object type we're displaying. Used for iteration.
    \param screen_id this screen's identifier.

    Checklist for derived constructors:
    - create all tiles and add them with addTile(). You should include a
      WControlScreenHeaderTile.
    - create starchart widget and hand it to addChart() */
WControlScreen::WControlScreen(GObjectType& type, WScreenId::ScreenId screen_id)
    : screen_id(screen_id),
      type(type),
      conn_pos_changed(type.sig_pos_changed.add(this, &WControlScreen::onPosChanged)),
      scanner(),
      chart(0),
      console_request(false)
{
    setSkin(holder.add(new Skin()));
    setExtent(GfxRect(*ui_screen), true);
    setGeometry();
    setState(st_Modal, true);

    const int scanner_height = 2*font_heights[FONT_NORMAL];
    add(scanner, GfxRect(chart_x1+2,                chart_y2 + 5,
                         extent.w - chart_x1 - 5-2, scanner_height));
    add(holder.add(new UIPrefixArg()), GfxRect(0, 0, 0, 0));
    add(holder.add(new WSelectionSwitcher()), GfxRect(0, 0, 0, 0));

    pushHistoryScreenId(WScreenId(screen_id, type.getCurrent()));
}

WControlScreen::~WControlScreen()
{ }

/** Compute geometry parameters. */
void
WControlScreen::setGeometry()
{
    tile_width = 315;           // FIXME: should probably be related to font size somehow
    tile_new_y = extent.h - 8*font_heights[FONT_NORMAL];
    tile_y     = 5;
    tile_x     = 5;
    chart_y2   = tile_new_y - 3*font_heights[FONT_NORMAL];
    chart_x1   = tile_width + tile_x + 5;
}

/** Selection-change handler. This is called when the user picks a
    different object on this object's type. */
void
WControlScreen::onPosChanged(int /*id*/)
{
    pushHistoryScreenId(WScreenId(screen_id, type.getCurrent()));
    update();
}

/** Add starchart widget. Must be called once. */
void
WControlScreen::addChart(WScannerChartWidget& chart)
{
    this->chart = &chart;
    add(chart, GfxRect(chart_x1, 5, extent.w - chart_x1 - 5, chart_y2 - 5));
    scanner.setChart(chart);
}

/** Add a tile. */
void
WControlScreen::addNewTile(WTile* t)
{
    /* Remember it */
    WTile& tile = holder.add(t);

    /* Add it */
    UIWidget::LayoutInfo info;
    tile.getLayoutInfo(info);

    if (tile_y + info.pref_size.y > extent.h - 5) {
        /* tile does not fit in this column. Would it fit when we
           remove the heading? */
        if (tile.getDisplayState() == WTile::sNormal) {
            tile.setDisplayState(WTile::sContentOnly);
            tile.getLayoutInfo(info);
            if (tile_y + info.pref_size.y <= extent.h - 5)
                goto success;
            tile.setDisplayState(WTile::sNormal);
            tile.getLayoutInfo(info);
        }

        /* start new column */
        tile_y = tile_new_y;
        tile_x += tile_width;

        /* does it fit there? */
        if (tile_y + info.pref_size.y > extent.h - 5) {
            /* nope. Remove heading if there, but we cannot do
               anything else so let it be. */
            if (tile.getDisplayState() == WTile::sNormal) {
                tile.setDisplayState(WTile::sContentOnly);
                tile.getLayoutInfo(info);
            }
        }
    }
 success:
    GfxRect tile_size(tile_x, tile_y, tile_width, info.pref_size.y);
    tile.setExtent(tile_size, true);
    tile_y += info.pref_size.y + 2;

    add(tile, tile_size);
}

/** Define keymap for this screen. */
void
WControlScreen::setKeymap(IntKeymapRef keymap)
{
    this->keymap = keymap;
}

void
WControlScreen::drawContent(GfxCanvas& can)
{
    // FIXME: optimize!
    GfxContext ctx;
    ctx.useColorScheme(getSkin());
    ctx.useCanvas(can);
    drawBackground(ctx, extent);

#ifdef BLUESKIN
    ctx.setRawColor(standard_colors[FCOLOR]);
    drawRectangle(ctx, GfxRect(extent.x, extent.y+1, extent.w, extent.h-2));
    drawRectangle(ctx, GfxRect(extent.x+1, extent.y, extent.w-2, extent.h));

    drawRectangle(ctx, GfxRect(chart_x1-1, 5-2,        extent.w - chart_x1 - 3, chart_y2 - 5+4));
    drawRectangle(ctx, GfxRect(chart_x1-2, 5-1,        extent.w - chart_x1 - 1, chart_y2 - 5+2));

    drawRectangle(ctx, GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#else
    drawFrameUp(can, extent);
    drawFrameDown(can, GfxRect(chart_x1-1, 5-1,        extent.w - chart_x1 - 3, chart_y2 - 5+2));
    drawFrameDown(can, GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#endif
    can.update(extent);
}

bool
WControlScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        /* Check user-defined keymaps */
        if (keymap != 0) {
            uint32_t command = keymap->lookupCommand(e.key.code);
            if (isAtom(command)) {
                IntTemporaryContextLock lock;
                setScriptContext();
                executeStatement(atomStr(command), e.key.arg);
                checkScriptRequest();
                return true;
            }
        }
        /* Check hardwired commands */
        switch(e.key.code) {
         case 'y':
            if (chart->isValid())
                chart->doKeyboardMode();
            return true;
        }
    }
    return UIFixedWidget::handleEvent(e, second_pass);
}

/** Browse through objects. */
void
WControlScreen::doBrowse(bool forward, bool marked)
{
    int n;
    if (forward)
        n = type.getNextWrap(type.getCurrent(), marked);
    else
        n = type.getPreviousWrap(type.getCurrent(), marked);

    if (type.isExistant(n))
        type.setCurrent(n);
}

/** Get scanner position. */
GPoint
WControlScreen::getScannerPosition() const
{
    return chart->getScannerPosition();
}

/** Run this control screen.
    \return Id of screen to activate (1..3 for ship/planet/base, 0 for race screen, etc.) */
void
WControlScreen::runScreen()
{
    // Initialize
    update();
    ui_root->add(*this);

    // Initial script run
    runRunnableProcesses();
    killTerminatedProcesses();
    checkScriptRequest();

    // Execute
    UIFixedWidget::run();
    ui_root->close(*this);
}

int
WControlScreen::getProperty(IntUserInterfaceProperty p)
{
    switch (p) {
     case iuiScreenNumber:
        return screen_id;
     case iuiScreenRegistered:
        return 1;
     case iuiChartX:
        if (chart->isValid())
            return chart->getCenterPosition().x;
        else
            return -1;
     case iuiChartY:
        if (chart->isValid())
            return chart->getCenterPosition().y;
        else
            return -1;
     case iuiScanX:
        if (chart->isValid())
            return chart->getScannerPosition().x;
        else
            return -1;
     case iuiScanY:
        if (chart->isValid())
            return chart->getScannerPosition().y;
        else
            return -1;
     case iuiSimFlag:
        return 0;
    }
    return -1;
}

void
WControlScreen::doGoToScreen(int screen, int obj)
{
    if (screen == 0) {
        stop(0);
    } else if (screen == 4) {
        GPoint pt = chart->isValid() ? chart->getCenterPosition() : getStarchartPosition();
        doGoToChart(pt);
    } else {
        stop(0);
        getPreviousBinding()->doGoToScreen(screen, obj);
    }
}

void
WControlScreen::doGoToChart(GPoint pt)
{
    /* Special magic when switching to the starchart position we are at:
       lock current object */
    if (chart->isValid() && pt == chart->getCenterPosition()) {
        setStarchartObject(type.getObject(type.getCurrent()));
    }

    stop(0);
    getPreviousBinding()->doGoToChart(pt);
}

void
WControlScreen::doMoveScanner(GPoint pt)
{
    if (chart->isValid())
        chart->setScannerPosition(pt);
}

void
WControlScreen::doPopupConsole()
{
    console_request = true;
}

void
WControlScreen::doOverlayMessage(string_t msg)
{
    showDecayingMessage(msg);
}

void
WControlScreen::doLeaveRace()
{
    stop(0);
    getPreviousBinding()->doLeaveRace();
}

void
WControlScreen::doSetChartView(string_t /*name*/)
{
    throw IntError::contextError();
}

IntKeymapRef
WControlScreen::getKeymap()
{
    return keymap;
}

void
WControlScreen::checkScriptRequest()
{
    while (console_request) {
        console_request = false;
        IntTemporaryContextLock lock;
        setScriptContext();
        doConsole();
    }
}
