/**
  *  \file client/scr-env.cc
  *  \brief Environment Scan Screen ("Mining Survey")
  *
  *  This is the "F5" screen. It displays information about a planet
  *  and thus is the main interface to the planet history.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/global.h"
#include "game/host.h"
#include "game/planet.h"
#include "game/tables.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/nls.h"
#include "game/formula.h"
#include "game/ship.h"
#include "game/team.h"
#include "game/racename.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/widgets/statictext.h"
#include "client/tiles/planetmining.h"
#include "ui/widgets/button.h"

/*
      Neutronium
      abundand, large masses
      10000kt [######-------]
      10000kt [######.......]
      99%     [###..........]

 */

struct UnloadInformation {
    int32 hostile_unload;
    int32 friendly_unload;
    bool  hostile_unload_is_IA;
    bool  hostile_unload_is_assumed;

    UnloadInformation()
        : hostile_unload(0),
          friendly_unload(0),
          hostile_unload_is_IA(false),
          hostile_unload_is_assumed(false)
        { }
};

static inline int32
pdiv(int32 a, int32 b)
{
    if (host.isPHost())
        return a/b;
    else
        return rdivaddl(a, b, 0);
}

static int
computeBaseTax(int player, int nid, int spi, int32 natives, int est_change)
{
    const int mifa = 100;
    int tax = host.getNativeTaxRateLimit(player);
    while (tax > 0) {
        /* Host 3.22.040 formula:
           Trunc(1000.0 - sqrt(Planet.Natives) - 85 * Planet.NativeTax
                        - (Planet.Mines + .Factories) DIV 2
                        - 50*(10-Planet.NativeGov))
           DIV 100
           change:=5 + Trunc(spi/2 - (Sqrt(Natives) + 85*LONGINT(tax) + 100/2) / 100);
           change:=5 + Trunc(50*spi - (85*LONGINT(tax) + 50) - Sqrt(Natives)) DIV 100;
           since MIFA is even, the same formula can be used for PHost and THost. */
        int change = int32(500 - (mifa/2) + (50*spi) - (85*tax) - std::sqrt(double(natives))) / 100;
        if (nid == nr_Avian)
            change += 10;
        if (change >= est_change)
            break;
        --tax;
    }
    return tax;
}

static string_t
formatBaseTax(const GPlanet& p, string_t name, int est_change)
{
    // Compute tax
    int tax = computeBaseTax(getPlayerId(), p.getNativeRace(), p.getNativeGovernment(), p.getNatives(), est_change);

    // Compute due
    int due = p.getNativeDue(tax, p.getNativeRace(), p.getNativeGovernment(), p.getNatives(), getPlayerId());

    // How many colonists needed to collect that?
    int rate = config.NativeTaxRate(p.getOwner());
    if (p.getNativeRace() == nr_Insectoid)
        rate *= 2;
    int colonists = (rate > 0 ? due * 100 / rate : due);

    // Show info
    if (colonists == due)
        return format(_("%s: %d%% (%d mc)"), name, tax, due);
    else
        return format(_("%s: %d%% (%d mc, with %d colonist clan%!1{s%})"), name, tax, due, colonists);
}



namespace {
    class WPlanetInfoBox : public UIWidget {
        GPlanet& planet;
        int draw_y;
        int font;
        UnloadInformation& unload;

        void write(GfxContext& ctx, int x, string_t text);
        void showSupport(GfxContext& ctx, int player);
        void showAge(GfxContext& ctx, int turn);
        void showBeamDown(GfxContext& ctx);
     public:
        WPlanetInfoBox(GPlanet& planet, UnloadInformation& unload);
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& li);
    };
}

WPlanetInfoBox::WPlanetInfoBox(GPlanet& planet, UnloadInformation& unload)
    : planet(planet), unload(unload)
{ }

void
WPlanetInfoBox::write(GfxContext& ctx, int x, string_t text)
{
    outTextF(ctx, extent.x + x, extent.y + draw_y, extent.w - x, text);
    draw_y += ctx.getFont().getTextHeight(text);
}

void
WPlanetInfoBox::showSupport(GfxContext& ctx, int player)
{
    write(ctx, 5, format(_("Supports %s %ss"),
                         clansToString(planet.getMaxSupportedColonists()),
                         player_racenames.getAdjName(player)));
    // FIXME 'won't die ...'
}

void
WPlanetInfoBox::showAge(GfxContext& ctx, int turn)
{
    if (turn > 0) {
        int age = getDisplayedTurn().getTurnNumber() - turn;

        if (age >= 3)
            ctx.setColor(UIColor::tc_Red);
        else
            ctx.setColor(UIColor::tc_Static);

        if (age == 0)
            write(ctx, 25, _("(current turn)"));
        else if (age == 1)
            write(ctx, 25, _("(previous turn)"));
        else if (age < 0)
            write(ctx, 25, format(_("(in %d turn%!1{s%})"), -age));
        else
            write(ctx, 25, format(_("(%d turns ago)"), age));
    }
}

void
WPlanetInfoBox::showBeamDown(GfxContext& ctx)
{
    /* What are we beaming down? */
    ctx.setColor(UIColor::tc_Static);
    write(ctx, 5, format(translate(unload.hostile_unload_is_assumed
                                   ? N_("Assumed, we'd beam down %d clan%!1{s%}")
                                   : N_("Beaming down %d clan%!1{s%}")),
                         unload.hostile_unload));

    /* Will that be a fight after all? */
    if (!planet.isOwnerKnown() || planet.getOwner() == 0 || planet.getOwner() == getPlayerId())
        return;

    /* Do we know how many enemies are there? */
    if (!planet.isCargoAmountKnown(el_Colonists) || planet.getCargo(el_Colonists) == 0)
        return;

    /* Okay, there will be ground combat. */
    int32 theirClans   = planet.getCargo(el_Colonists);
    int16 theirDefense = planet.getMaxStructures(str_Defense, theirClans);

    // Minimum power: assuming 0 defense
    int32 theirRatioLo = config.GroundDefenseFactor(planet.getOwner());
    int32 theirPowerLo = theirRatioLo * theirClans;

    // Maximum power: assuming full defense
    int32 theirPowerHi;
    int32 theirRatioHiX20;
    if (host.isPHost()) {
        theirRatioHiX20 = 20*theirRatioLo + theirDefense;
        theirPowerHi    = (theirClans * theirRatioHiX20) / 20;
    } else {
        theirRatioHiX20 = theirRatioLo * (theirDefense + 20);
        theirPowerHi    = rdivaddl(theirClans * theirRatioHiX20, 20, 0);
    }

    // Compute my power
    int32 myRatio = config.GroundKillFactor(getPlayerId());
    int32 myPower = myRatio * unload.hostile_unload;

    // Avoid crash on bogus config
    if (myRatio == 0 || theirRatioLo == 0)
        return;

    if (myPower > theirPowerHi) {
        // Guaranteed win
        write(ctx, 5, (unload.hostile_unload_is_assumed
                       ? _("We'd win this ground combat.")
                       : _("We'll win this ground combat.")));

        uint32 a = pdiv(myPower - theirPowerLo, myRatio);
        uint32 b = pdiv(myPower - theirPowerHi, myRatio);
        if (a == b)
            write(ctx, 5, format(_("%d clan%1{ survives%|s survive%}"), a));
        else
            write(ctx, 5, format(_("%d to %d clans survive"), a, b));
    } else if (myPower < theirPowerLo) {
        // Guaranteed loss
        ctx.setColor(UIColor::tc_Red);
        write(ctx, 5, (unload.hostile_unload_is_assumed
                       ? _("We'd lose this ground combat.")
                       : _("We'll lose this ground combat.")));

        uint32 a = pdiv(theirPowerLo - myPower, theirRatioLo);
        uint32 b = pdiv(20*(theirPowerHi - myPower), theirRatioHiX20);
        if (a == b)
            write(ctx, 5, format(_("%d of their clans survive%1{s%}"), a));
        else
            write(ctx, 5, format(_("%d to %d of their clans survive"), a, b));
    } else {
        // Depends on defense
        ctx.setColor(UIColor::tc_Yellow);
        write(ctx, 5, format(_("Chance to win ground combat: %d%%"),
                             (99*(myPower - theirPowerLo+1)) / (theirPowerHi - theirPowerLo+1)));

        uint32 a = pdiv(myPower - theirPowerLo, myRatio);
        uint32 b = pdiv(20*(theirPowerHi - myPower), theirRatioHiX20);
        write(ctx, 5, format(_("Up to %d of our clans survive%1{s%}"), a));
        write(ctx, 5, format(_("Up to %d of their clans survive%1{s%}"), b));
    }
}

void
WPlanetInfoBox::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setSolidBackground();

    draw_y = 0;
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(UIColor::tc_Heading);
    write(ctx, 0, _("Climate"));

    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);

    mp16_t temp = planet.getTemperature();
    if (temp.isKnown()) {
        write(ctx, 5, format(_("Climate type: %s"), getTemperatureName(temp)));
        write(ctx, 5, format(_("Average temperature: %d" "\xC2\xB0" "F"), temp));
        if (planet.isOwnerKnown() && planet.getOwner() > 0 && planet.getOwner() != ::getPlayerId())
            showSupport(ctx, planet.getOwner());
        showSupport(ctx, ::getPlayerId());
    } else {
        // FIXME: use word-wrapping functions
        write(ctx, 5, _("(no information on"));
        write(ctx, 5, _("climate available)"));
    }

    draw_y += 10;
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(UIColor::tc_Heading);
    write(ctx, 0, _("Population"));
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);

    if (!planet.getNativeRace().isKnown()) {
        if (planet.getNatives().isKnown()) {
            write(ctx, 5, _("Unknown native race"));
            write(ctx, 5, format(_("Population: %s"), clansToString(planet.getNatives())));
        } else if (planet.isKnownToHaveNatives()) {
            write(ctx, 5, _("Planet has natives"));
        } else {
            write(ctx, 5, _("No information on natives available"));
        }
    } else if (planet.getNativeRace() == 0) {
        write(ctx, 5, _("No native population"));
    } else {
        write(ctx, 5, format(_("Native race: %s"),
                             getNativeRaceName(planet.getNativeRace())));
        if (planet.getNatives().isKnown())
            write(ctx, 5, format(_("Population: %s"), clansToString(planet.getNatives())));
        if (planet.getNativeGovernment().isKnown())
            write(ctx, 5, format(_("Government: %s"),
                                 getNativeGovernmentName(planet.getNativeGovernment())));
    }
    if (!planet.hasPlanetData()) {
        showAge(ctx, planet.getHistoryTimestamp(ts_Natives));
    }

    /* Show taxes */
    if (planet.getNativeRace().isKnown()
        && planet.getNatives().isKnown()
        && planet.getNativeGovernment().isKnown()
        && planet.getNativeRace() > 0
        && planet.getNativeRace() != nr_Amorphous)
    {
        write(ctx, 5, formatBaseTax(planet, _("Base Tax Rate"), 0));
        write(ctx, 5, formatBaseTax(planet, _("Max Tax Rate"), -30));
    }
    // FIXME: assimilation forecast

    ctx.setColor(UIColor::tc_Static);
    draw_y += 3;
    if (!planet.isOwnerKnown()) {
        write(ctx, 5, _("No information on colonists available"));
    } else if (planet.getOwner() == 0) {
        write(ctx, 5, _("No colonists"));
    } else {
        write(ctx, 5, format(_("Colonists: %s"), player_racenames.getAdjName(planet.getOwner())));
        if (planet.getColonists().isKnown())
            write(ctx, 5, format(_("Population: %s"), clansToString(planet.getColonists())));
    }

    /* Industry */
    string_t s;
    mp16_t fact = planet.getStructures(str_Factories);
    mp16_t mine = planet.getStructures(str_Mines);
    mp16_t defe = planet.getStructures(str_Defense);
    if (!fact.isKnown() && !mine.isKnown()) {
        mp16_t level = planet.getIndustryLevel();
        if (level.isKnown())
            write(ctx, 5, format(_("%s industrial activity"), getIndustryLevelName(level)));
    } else {
        if (fact.isKnown() && fact > 0)
            s += format(_(", %d factor%!1{ies%|y%}"), fact);
        if (mine.isKnown() && mine > 0)
            s += format(_(", %d mine%!1{s%}"), mine);
        if (defe.isKnown() && defe > 0)
            s += format(_(", %d DP%!1{s%}"), defe);
        if (!s.empty())
            write(ctx, 5, s.substr(2));
    }
    if (!planet.hasPlanetData()) {
        showAge(ctx, planet.getHistoryTimestamp(ts_Colonists));
    }

    /* Cash etc. */
    ctx.setColor(UIColor::tc_Static);
    s.clear();
    if (planet.isCargoAmountKnown(el_Money))
        s += format(_(", %d mc"), planet.getCargo(el_Money));
    if (planet.isCargoAmountKnown(el_Supplies))
        s += format(_(", %d suppl%1{y%|ies%}"), planet.getCargo(el_Supplies));
    if (!s.empty()) {
        write(ctx, 5, s.substr(2));
        if (!planet.hasPlanetData()) {
            showAge(ctx, planet.getHistoryTimestamp(ts_Cash));
        }
    }

    /* FCode */
    ctx.setColor(UIColor::tc_Static);
    if (planet.isFCodeKnown()) {
        s = planet.getFCode();
        if (planet.hasPlanetData())
            write(ctx, 5, format(_("Friendly code: %s"), s));
        else
            write(ctx, 5, format(_("Last known friendly code: %s"), s));
    }
    if (planet.hasBase())
        write(ctx, 5, _("Starbase present"));
    else if (planet.isBuildingBase())
        write(ctx, 5, _("Starbase being built"));

    /* Unload information */
    if (unload.friendly_unload != 0) {
        draw_y += 4;
        ctx.setColor(UIColor::tc_Yellow);
        write(ctx, 5, format(_("Friendly unload: %d clan%!1{s%}"), unload.friendly_unload));
    }
    if (unload.hostile_unload != 0) {
        if (unload.friendly_unload == 0)
            draw_y += 4;
        if (unload.hostile_unload_is_IA) {
            ctx.setColor(UIColor::tc_Red);
            write(ctx, 5, _("Imperial Assault!"));
        }
        showBeamDown(ctx);
    }

    can.update(getExtent());
}

/* Display on info box:
    <big> Climate
    Climate type: Temp               or: (no information available)
    Average temperature: 99F
    <support for owner>
    <support for us>

    <big>Population
    no natives
      or: Native race: foo
          Population:  100c
          Government:  bar
          Assimilated in...
          (age)
    no information on colonists
      or: Colonists: race
          Population: ...
          ... industrial activity
          or: n factories, n defense, n mines
          (age)
          n mc, n supplies
          (age)
          Friendly code: ...
          or: Last known fcode: ...
          Starbase present/being built
          beaming down NN clans, imperial assault */

void
WPlanetInfoBox::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size = GfxPoint(310, 25 * font_heights[FONT_NORMAL]);
}

static void
computeUnload(UnloadInformation& info, const GUniverse& univ, const GPlanet& pl)
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        const GShip& sh = univ.getShip(sid);
        if (sh.isExistant()
            && sh.getPos() == pl.getPos()
            && sh.isPlayed()
            && sh.getOwner() == getPlayerId()
            && sh.isTransporterActive(GShip::t_Unload))
        {
            /* ship exits, is at this planet, played by current player, and unloading */
            if (sh.getRealOwner() == pl.getOwner()) {
                /* ship actually belongs to planet owner, so there will not be a fight */
                info.friendly_unload += sh.getTransporter(GShip::t_Unload).colonists;
            } else {
                /* hostile unload */
                info.hostile_unload += sh.getTransporter(GShip::t_Unload).colonists;
                if (sh.canDoSpecial(hf_ImperialAssault) && sh.getTransporter(GShip::t_Unload).colonists >= 10)
                    info.hostile_unload_is_IA = true;
            }
        }
    }
}

void
doPlanetScan(int pid)
{
    static const UIKeyTranslator::Item keys[] = {
        { ' ',          1, true },
        { SDLK_ESCAPE,  1, true },
        { SDLK_F5,      1, true },
    };
    GPlanet& p = getDisplayedTurn().getCurrentUniverse().getPlanet(pid);
    if (!p.hasAnyPlanetData()) {
        messageBox(format(_("No information available about planet %s (#%d)."),
                          p.getName(), pid),
                   _("Planet Scan"));
        return;
    }

    UnloadInformation unload;
    computeUnload(unload, getDisplayedTurn().getCurrentUniverse(), p);

    ObjectHolder h;
    UIWindow win(p.getName(), UIWindow::BLUE_WINDOW, COLOR_DARK);
    win.setLayout(UIHBoxLayout::instance5);
    UIGroup lgroup(UIVBoxLayout::instance5), rgroup(UIVBoxLayout::instance5),
        bgroup(UIHBoxLayout::instance5);
    WPlanetInfoBox info(p, unload);

    bgroup.add(h.add(new UIButton(1, _("Close"), SDLK_RETURN)).addStop());
    bgroup.add(h.add(new UISpacer()));
    bgroup.add(h.add(new UIKeyTranslator(keys, countof(keys))));
    lgroup.add(info);
    lgroup.add(bgroup);
    if (p.getHistoryTimestamp(ts_Minerals) > 0) {
        // FIXME: display age of this information
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Neutronium)));
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Tritanium)));
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Duranium)));
        rgroup.add(h.add(new WPlanetMiningTile(p, el_Molybdenum)));
    } else {
        rgroup.add(h.add(new UIStaticText(_("Minerals"), UIStaticText::WhiteTitle)));
        rgroup.add(h.add(new UIMultilineStatic(FONT_NORMAL,
                                               _("(no information on minerals available)"),
                                               295)));
        rgroup.add(h.add(new UISpacer()));
    }
    win.add(lgroup);
    win.add(rgroup);
    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    ui_root->run();
    ui_root->close(win);
}
