/**
  *  \file client/screens/unpackscreen.cc
  *  \brief Unpack screen
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/unpackscreen.h"
#include "client/dialogs/consolelog.h"
#include "client/widgets/setselect.h"
#include "cpluslib/format.h"
#include "game/gamedirs.h"
#include "game/load.h"
#include "game/rst.h"
#include "game/struct.h"
#include "game/unpack.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/multilinestatic.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"
#include "client/widgets/imagepanel.h"
#include "ui/widgets/alignwidget.h"
#include "resmgr/resmgr.h"
#include "ui/widgets/button.h"

namespace {
    enum { WIDTH = 600 };

    class WUnpackWindow : public WImagePanel {
        ObjectHolder h;
        WPlayerSetSelector& set_sel;

        void init();
     public:
        WUnpackWindow(WPlayerSetSelector& set_sel);
        ~WUnpackWindow();

        void onOK();
    };

    class AskingAnswerProvider : public AnswerProvider {
     public:
        bool ask(int, string_t q)
            {
                return UIMessageBox(q, _("Unpack")).doYesNoDialog();
            }
    };
}

WUnpackWindow::WUnpackWindow(WPlayerSetSelector& set_sel)
    : WImagePanel(_("Unpack"), requestPixmap(ResId("bg.unpack", ui_root->getExtent().w))),
      set_sel(set_sel)
{
    init();
}

WUnpackWindow::~WUnpackWindow()
{ }

void
WUnpackWindow::init()
{
    UIGroup& main = h.add(new UIGroup(UIVBoxLayout::instance5));

    main.add(h.add(new UIMultilineStatic(FONT_NORMAL, _("Choose files to unpack:"), WIDTH)));
    main.add(set_sel);
    main.add(h.add(new UIMultilineStatic(FONT_NORMAL, _("The Unpack function will unpack result files into "
                                                        "individual files. After unpacking, you can play your "
                                                        "turn using all VGAP utilities of your choice. When "
                                                        "done with the turn, use the Maketurn function to "
                                                        "generate TRN files to send to the host.\n"
                                                        "Note that PCC2 doesn't require you to unpack a result "
                                                        "file prior to playing, but most other utilities do."),
                                         WIDTH)));

    UIButton& btn_ok     = h.add(new UIButton(1, _("OK"), SDLK_RETURN));
    UIButton& btn_cancel = h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE));

    btn_ok.sig_fired.add(this, &WUnpackWindow::onOK);
    btn_cancel.addStop();

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(btn_ok);
    g.add(btn_cancel);
    g.add(h.add(new UISpacer()));
    main.add(g);

    add(h.add(new UIAlignedWidget(main, GfxPoint(1, 1), GfxPoint(10, 10))));
    setExtent(ui_root->getExtent());
}

void
WUnpackWindow::onOK()
{
    // Fetch selection
    GPlayerSet selected = set_sel.getSelectedPlayers();
    if (selected.empty())
        return;

    // Show console. First, determine size. Default is 7 lines per player.
    int nlines = 0;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (selected.contains(i) && nlines < 14)
            nlines += 7;
    WConsoleLogWindow log(_("Unpack"), nlines);
    log.scrollAfterEnd();
    log.open();

    // Unpack selection
    AskingAnswerProvider aap;
    GUnpacker theUnpacker;
    theUnpacker.setFormat(theUnpacker.WindowsFormat);
    theUnpacker.setAnswerProvider(aap);
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (selected.contains(i)) {
            string_t filename = format("player%d.rst", i);
            try {
                Ptr<Stream> rst = game_file_dir->openFileNT(filename, Stream::C_READ);
                GResultFile rstfile(*rst);
                console.write(LOG_INFO, format(_("=== Unpacking player %d... ==="), i));
                theUnpacker.unpack(rstfile, i);
            }
            catch (FileProblemException& e) {
                console << e.getFileName() << ": " << e.what() << "\n";
            }
            catch (std::exception& e) {
                console << filename << ": " << e.what() << "\n";
            }
        }
    }
    log.waitForConfirmation();
    stop(1);
}    

/** Do unpack screen. Asks the user for a set of players in the
    game_file_dir, and unpacks their results.
    \param defaultSet players selected by default */
void
doUnpack(GPlayerSet defaultSet)
{
    // VBox
    //   UIMultilineStatic
    //   WPlayerSetSelector
    //   HBox
    //     UIButton "OK"
    //     UIButton "Cancel"

    /* Check what's available */
    GDirectoryOverview ov;
    ov.scan(*game_file_dir, true);
    GPlayerSet avail = ov.getPlayersWhere(ov.have_RST);
    if (avail.empty()) {
        messageBox(_("This directory contains no files to unpack."),
                   _("Unpack"));
        return;
    }

    /* Build player selector */
    WPlayerSetSelector psel(1000, NUM_PLAYERS, WIDTH);
    psel.setSelectablePlayers(avail);
    psel.setSelectedPlayers(avail & defaultSet);
    psel.scrollTo(0, 1);
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (ov.getPlayerFlags(i) & ov.have_RST)
            psel.setInfo(i, _("RST"));
    }

    /* Build window */
    WUnpackWindow win(psel);
    ui_root->centerWidget(win);
    ui_root->runModal(win);
}
