/**
  *  \file client/wcomplist.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wcomplist.h"
#include "game/planet.h"
#include "game/team.h"
#include "game/reg.h"

/*************************** WComponentListbox ***************************/

/** \class WComponentListbox
    \brief List box containing list of ship parts

    This class only provides a method drawComponent() to draw ship
    parts. Derived classes use this to achieve a uniform look across
    all component lists. */

/** Contructor.
    \param id       widget id
    \param lines    number of lines of widget
    \param width    width of widget
    \param planet   planet. May be null. Ignored if it doesn't have a
                    starbase. Tech levels (and tick marks) are taken
                    from here. */
WComponentListbox::WComponentListbox(int id, int lines, int width, GPlanet* planet)
    : UIListbox(id, lines, width, 0), tick_planet(planet)
{
    if (!tick_planet->hasBase())
        tick_planet = 0;
}

/** Draw a component entry.
    \param ctx     graphics context, prepared by prepareEntry()
    \param area    area, prepared by prepareEntry()
    \param com     component to draw.
    \param tech_area the tech area used by this component.
    \param have    amount of such components in storage (used for tick marks) */
void
WComponentListbox::drawComponent(GfxContext& ctx, GfxRect& area, const GComponent& com, TTechLevel tech_area, int have)
{
    if (tick_planet && com.getTech() > tick_planet->getBaseTech(tech_area)) {
        if (com.getTech() > 6 && !allowRegistered(tick_planet->getOwner()))
            ctx.setColor(UIColor::tc_Faded);
        else
            ctx.setColor(UIColor::tc_Blue);
    }
    const char* tick = have ? "\xE2\x80\xA2" : " ";
    outTextF(ctx, area.x, area.y, area.w, tick + com.getName());
}

/****************************** WHullListbox *****************************/

/** \class WHullListbox
    \brief Hull List box

    This widget displays a list of hulls. You can (actually, must)
    provide a list of hulls to display. */

/** Constructor. \see WComponentListbox::WComponentListbox() */
WHullListbox::WHullListbox(int id, int lines, int width, GPlanet* p)
    : WComponentListbox(id, lines, width, p)
{ }

WHullListbox::~WHullListbox()
{ }

void
WHullListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        const int slot = i + getTop();
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (slot < int(content.size())) {
            const int hull = content[slot];
            int have = 0;
            if (tick_planet && tick_planet->hasBase()) {
                int n = getTruehullSlot(tick_planet->getOwner(), hull);
                if (n)
                    have = tick_planet->getBaseHullStoreSlot(n);
            }
            drawComponent(ctx, area, getHull(hull), tl_Hull, have);
        }
    }
}

/** Add a hull to the end of the list.
    \param n hull number (1..NUM_HULLS)
    \see swapContents() */
void
WHullListbox::addHullId(int n)
{
    content.push_back(n);
    setSize(content.size());
    drawWidget();
}

/** Set content of this list.
    \param new_content vector of hull numbers (1..NUM_HULLS). After
    this function completes, this widget will display the old contents
    of /new_content/, and /new_content/ will contain the old content. */
void
WHullListbox::swapContents(std::vector<int>& new_content)
{
    int now = getCurrentHull();
    content.swap(new_content);
    setSize(content.size());
    scrollToHull(now);
    adjustPosition();
    drawWidget();
}

/** Get currently-selected hull number. */
int
WHullListbox::getCurrentHull() const
{
    if (getCurrent() < int(content.size()))
        return content[getCurrent()];
    else
        return 0;
}

/** Scroll to specified hull. Does nothing if this hull is not on our
    list. */
void
WHullListbox::scrollToHull(int id)
{
    for (std::vector<int>::size_type i = 0; i < content.size(); ++i) {
        if (id == content[i]) {
            scrollTo(i, 1);
            return;
        }
    }
}

/************ WEngineListbox / WLauncherListbox / WBeamListbox ***********/

WEngineListbox::WEngineListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_ENGINES, width, tick_planet)
{
    setSize(NUM_ENGINES);
}

void
WEngineListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (slot <= NUM_ENGINES)
            drawComponent(ctx, area, getEngine(slot), tl_Engine,
                          tick_planet ? tick_planet->getBaseEngineStore(slot) : 0);
    }
}

WBeamListbox::WBeamListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_BEAMS, width, tick_planet)
{
    setSize(NUM_BEAMS);
}

void
WBeamListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (slot <= NUM_BEAMS)
            drawComponent(ctx, area, getBeam(slot), tl_Beam,
                          tick_planet ? tick_planet->getBaseBeamStore(slot) : 0);
    }
}

WLauncherListbox::WLauncherListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_TORPS, width, tick_planet)
{
    setSize(NUM_TORPS);
}

void
WLauncherListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (slot <= NUM_TORPS)
            drawComponent(ctx, area, getTorp(slot), tl_Torpedo,
                          tick_planet ? tick_planet->getBaseLauncherStore(slot) : 0);
    }
}
