/**
  *  \file game/objlist.h
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJLIST_H
#define PCC_V2_GAME_OBJLIST_H

#include <vector>
#include "cpluslib/types.h"
#include "game/objtypes.h"
#include "game/obj.h"
#include "game/objnote.h"

/** Flags for ship list and GObjectList::addObjectsAt(). */
enum {
    /* addObjectsAt */
    ls_IncludeForeignShips = 1,
    ls_IncludePlanet       = 2,
    ls_SafeShipsOnly       = 4,

    /* ship list */
    ls_AllowForeignShips   = 8,
    ls_AllowPlanet         = 16,
    ls_EarlyExit           = 32
};

/** Sort orders. */
enum {
    SortById          = 0,
    SortByOwner       = 1,
    SortByHull        = 2,
    SortByMass        = 3,
    SortByFleet       = 4,
    SortByTowGroup    = 5,
    SortByBattleOrder = 6,
    SortByLocation    = 7,
    SortByHullMass    = 8,
    SortByDamage      = 9,
    SortByName        = 10
};

/** Object reference. Identifies an object by its type and Id (=index into type).
    The object is therefore correctly identified during turn changes. The object
    reference may also be to a (currently) non-existant object. */
class GObjectId {
    GObjectType* type;
    int          id;
 public:
    GObjectId();
    GObjectId(GObjectType& type, int id);

    int          getId() const;
    string_t     getName() const;
    bool         isValid() const;
    GObject&     getObject() const;
    GObjectType& getType() const;
    int          getIndex() const;
    bool         isGameObject() const;
    bool         isExistant() const;

    bool operator==(const GObjectId& o) const;
    bool operator!=(const GObjectId& o) const;
};

/** Object list. Container for a list of objects (GObjectId's). It offers a number
    of functions to add objects and bring them into a form convenient to the user.
    It also contains a GNoteContainer to allow adding additional selectable items
    and dividers. */
class GObjectList : public GObjectType {
 public:
    typedef std::vector<GObjectId> container_t;
    typedef container_t::const_iterator iterator;
    typedef container_t::size_type size_type;

    // Construction:
    GObjectList();
    ~GObjectList();

    // Adding things:
    void addExistingObjects(GObjectType& t);
    void addObject(GObjectId id);
    void addObject(GObjectType& type, int id);
    void addObjectsAt(GUniverse& univ, GPoint pt, int flags, int exclude_ship);
    void addNonObject(int id, string_t name);

    // Manipulation:
    void sort(int (*func)(GObject&,GObject&));
    void sort(int32 (*func)(GObject&));
    void clear();
    void addDividers(string_t (*func)(GObject&));
    void removeDividers();
    void removeNonexistant();
    void postprocessUser(bool with_dividers);

    // Container access: (0-based)
    iterator begin() const
        { return list.begin(); }
    iterator end() const
        { return list.end(); }

    // GObjectType: (1-based!)
    void         init();
    bool         isExistant(int index);
    GObject&     getObject(int index);
    GObjectId    getObjectId(int index) const;
    int          getNextIndex(int index);
    int          getPreviousIndex(int index);
    int          getIndexFor(GObjectId id) const;
    int          getIndexFor(const GObject& obj) const;

    size_type    getNumObjects() const
        { return list.size(); }
    size_type    getNumGameObjects() const;
    size_type    getNumExistingObjects() const;

 private:
    container_t    list;
    GNoteContainer notes;
};

int32 sortById(GObject& o);
int32 sortByOwner(GObject& o);
int   sortByName(GObject& l, GObject& r);
int32 sortByHull(GObject& o);
int32 sortByMass(GObject& o);
int32 sortByHullMass(GObject& o);
int32 sortByDamage(GObject& o);

string_t diviOwner(GObject& o);
string_t diviHull(GObject& o);

int  getSortOrderConfig();
void setSortOrderConfig(int n);

/**************************** Inline Functions ***************************/

/** Construct blank object reference.
    \post !isValid() */
inline
GObjectId::GObjectId()
    : type(0), id(0)
{ }

/** Get Id of referred object. Note that this need not be the same as the
    index of the object, see comment for GObjectType.
    \pre isValid() */
inline int
GObjectId::getId() const
{
    return type->getObject(id).getId();
}

/** Get name of referred object.
    \pre isValid() */
inline string_t
GObjectId::getName() const
{
    return type->getObject(id).getFullName();
}

/** Check validity.
    \return true iff this GObjectId has been assigned a type/index pair. */
inline bool
GObjectId::isValid() const
{
    return type != 0;
}

/** Get referred object.
    \pre isValid() */
inline GObject&
GObjectId::getObject() const
{
    return type->getObject(id);
}

/** Get type of referred object. */
inline GObjectType&
GObjectId::getType() const
{
    return *type;
}

/** Get index of referred object. This is the index to pass into this
    object's type to receive the object. */
inline int
GObjectId::getIndex() const
{
    return id;
}

/** Check whether this is a game object. It must exist and not be part
    of a GNoteContainer. */
inline bool
GObjectId::isGameObject() const
{
    return isExistant() && dynamic_cast<GNoteContainer*>(type) == 0;
}

/** Check whether the referred object exists. */
inline bool
GObjectId::isExistant() const
{
    return type != 0 && type->isExistant(id);
}

/** Compare for equality. Object references are identical if they refer
    to the same object through the same type. */
inline bool
GObjectId::operator==(const GObjectId& o) const
{
    return type==o.type && id==o.id;
}

/** Compare for inequality. \see operator== */
inline bool
GObjectId::operator!=(const GObjectId& o) const
{
    return type!=o.type || id!=o.id;
}

#endif
