/**
  *  \file game/unitscore.cc
  *  \brief Per-Unit Scores
  *
  *  This manages per-unit scores; in particular, those are used for
  *  unit experience in PHost.
  *
  *  Although unit scores are generally stored indexed by type, then
  *  unit (i.e. a global list of score types, associated with a list
  *  of units' scores), we separate these two: a GUnitScoreDefinitions
  *  object defines all scores, and each object contains a list of
  *  applicable stores indexed by the GUnitScoreDefinitions, similar
  *  to interpreter properties split into IntVariableNames for
  *  indexing, and IntDataSegment with the actual content. This
  *  requires us to split up stuff we load, and gather it up again
  *  when we save it, but it allows us to easily clone an object with
  *  score and assign it a new score, either for loading past chart.cc
  *  files or for performing host updates.
  *
  *  It also needs a more memory (i.e. 1500 vector<Item>, many empty,
  *  instead of one or two definitions with one vector<Item> each),
  *  but this is not so much an issue today as it was in PCC 1.x.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/unitscore.h"
#include "cpluslib/assert.h"

/** Unit score definitions for ships. */
GUnitScoreDefinitions ship_score_definitions;

/** Unit score definitions for planets. */
GUnitScoreDefinitions planet_score_definitions;

/************************* GUnitScoreDefinitions *************************/

/** Unit score index that defines "no such score". */
const GUnitScoreDefinitions::index_t GUnitScoreDefinitions::nil;

/** Constructor. */
GUnitScoreDefinitions::GUnitScoreDefinitions()
{ }

/** Destructor. */
GUnitScoreDefinitions::~GUnitScoreDefinitions()
{ }

/** Add a score definition. Does nothing if the definition already exists,
    in this case this only returns the existing index.
    \param header Score definition
    \return index for the score definition, getScoreDefinition(return).score_id = header.score_id */
GUnitScoreDefinitions::index_t
GUnitScoreDefinitions::addScoreDefinition(const TUnitScoreHeader& header)
{
    index_t id = lookupScore(header.score_type);
    if (id == nil) {
        id = definitions.size();
        definitions.push_back(header);
    }
    return id;
}

/** Get score definition by index.
    \param index Index to look up, [0, getNumScores()).
    \return score definition */
const TUnitScoreHeader&
GUnitScoreDefinitions::getScoreDefinition(index_t index) const
{
    ASSERT(index < definitions.size());
    return definitions[index];
}

/** Look up score by identifier.
    \param score_type score_type field of the definition to look up
    \return index, or nil if none found */
GUnitScoreDefinitions::index_t
GUnitScoreDefinitions::lookupScore(int16 score_type) const
{
    for (index_t i = 0; i < definitions.size(); ++i)
        if (definitions[i].score_type == score_type)
            return i;
    return nil;
}

/** Get number of scores stored. */
GUnitScoreDefinitions::index_t
GUnitScoreDefinitions::getNumScores() const
{
    return definitions.size();
}

/** Clear (forget all definitions). */
void
GUnitScoreDefinitions::clear()
{
    defvec_t().swap(definitions);
}

/****************************** GUnitScores ******************************/

/** Constructor. Makes blank object. */
GUnitScores::GUnitScores()
{ }

/** Destructor. */
GUnitScores::~GUnitScores()
{ }

/** Set score value.
    \param index Score to set
    \param value Score value
    \param turn  Turn of this report */
void
GUnitScores::setScore(GUnitScoreDefinitions::index_t index, int16 value, int16 turn)
{
    ASSERT(index != GUnitScoreDefinitions::nil);
    static const Item null_item = { 0, -1 };
    if (items.size() <= index) {
        items.reserve(index+1);
        while (items.size() <= index)
            items.push_back(null_item);
    }
    items[index].turn  = turn;
    items[index].value = value;
}

/** Merge score value. Like setScore, but only modifies the score if this report's
    turn is the same or higher than the stored value.
    \param index Score to set
    \param value Score value
    \param turn  Turn of this report */
void
GUnitScores::mergeScore(GUnitScoreDefinitions::index_t index, int16 value, int16 turn)
{
    ASSERT(index != GUnitScoreDefinitions::nil);
    if (turn >= getTurn(index))
        setScore(index, value, turn);
}

/** Get turn for a score.
    \param index Score to read
    \return Turn number of associated score report, 0 if not known or never set */
int16
GUnitScores::getTurn(GUnitScoreDefinitions::index_t index) const
{
    if (index < items.size())
        return items[index].turn;
    else
        return 0;
}

/** Get score.
    \param index Score to read
    \return Score, -1 if not known or never set */
int16
GUnitScores::getScore(GUnitScoreDefinitions::index_t index) const
{
    if (index < items.size())
        return items[index].value;
    else
        return -1;
}

/******************************** Globals ********************************/

/** Initialize unit scores before loading a game. */
void
initUnitScores()
{
    planet_score_definitions.clear();
    ship_score_definitions.clear();
}

/** Shut down unit scores after exiting a game. */
void
doneUnitScores()
{
    initUnitScores();
}
